import itertools
import yaml
from pathlib import Path

# TODO: potentially subclass Path here - however, I've had a hard time doing this, may need to wait for an update
# would include read_yaml, get_name, and clean_workspace() as custom methods, change staticmethod to classmethod
class RPath:
    """Stores the location of core directories"""
    @staticmethod
    def texnew():
        return Path.home() / '.texnew'

    @staticmethod
    def workspace():
        return Path.home() / '.texnew' / '.workspace'

    @staticmethod
    def templates():
        return Path.home() / '.texnew' / 'templates'


def read_yaml(path):
    """Read a yaml file pointed to by a path."""
    return yaml.safe_load(path.read_text())


def clean_workspace():
    """Deletes all files in '~/.texnew/.workspace' except .gitignore"""
    for p in RPath.workspace().iterdir():
        if not p.name == ".gitignore":
            p.unlink()


def safe_rename(path):
    """Safely rename a file pointed to by path"""
    for t in itertools.count():
        new_path = path.with_name("{}_{}".format(path.stem, t) + "".join(path.suffixes))
        if not new_path.exists():
            path.rename(new_path)
            break



"""
TexNew

Automatic LaTeX template management.

:copyright: (c) 2019 by Alex Rutar
:license: MIT, see LICENSE for more details.
"""

__version__ = "1.6"

from .template import build, update
from .document import Document, Divider, TexnewDocument
from .rpath import RPath



import sys

from .template import build, update, load_template, load_user, available_templates
from .document import TexnewDocument
from .rpath import safe_rename, RPath
from pathlib import Path

def load_info(template_type):
    """Load template and user information simultaneously, with error handling."""
    try:
        data = load_template(template_type)
        if 'substitutions' not in data.keys():
            data['substitutions'] = {}
    except FileNotFoundError:
        print("Error: The template \"{}\" does not exist. The possible template names are:\n".format(template_type)+ "\t".join(available_templates()))
        sys.exit(1)

    try:
        user = load_user()
    except FileNotFoundError:
        print("Warning: no user file found.")
        user_info = {}

    data['substitutions'].update(user)
    return data


def run(fname, template_type):
    """Make a LaTeX file fname from template name template_type"""
    # load and catch basic errors with template choice
    fpath = Path(fname)
    if fpath.exists():
        print("Error: The file \"{}\" already exists. Please choose another filename.".format(fname))
        sys.exit(1)

    template_data = load_info(template_type)

    try:
        # update substitutions
        tdoc = build(template_data)
        tdoc.write(fpath)

    # TODO: this is pretty funny
    except Exception as e:
        print(e)
        sys.exit(1)


# TODO: add error handling here
def run_update(fname, template_type, transfer=['file-specific preamble', 'main document']):
    """Update given fname to new template_type, preserving blocks in transfer"""
    # basic checks
    fpath = Path(fname)
    if not fpath.exists():
        print("Error: No file named \"{}\" to update.".format(fname))
        sys.exit(1)

    # load the document
    tdoc = TexnewDocument.load(fpath)

    # generate replacement document
    template_data = load_info(template_type)
    new_tdoc = update(tdoc, template_data, transfer)

    # copy the existing file to a new location
    safe_rename(fpath)

    new_tdoc.write(fpath)


# TODO: catch if .workspace doesn't exist for some reason?
def run_test():
    """Compile and test every template"""
    for tm in available_templates():
        tdoc = build(load_info(tm))
        errors = tdoc.verify()
        if not errors:
            print("No errors in template '{}'".format(tm))
        else:
            print("Errors in template '{}'.".format(tm))



import re
import itertools
import subprocess

# TODO: clean .workspace instead of clean_dir
from .rpath import clean_workspace, RPath
from . import __version__
from pathlib import Path

class Divider:
    """Divider class primarily to be used in the document class"""
    def __init__(self, start_symb, fill, length=80):
        if len(fill) != 1:
            raise ValueError("'fill' must be a string of length 1")
        self.fill = fill
        self.start = start_symb
        self.length = length

    def gen(self, name):
        if len(name) >= self.length-len(self.start)-3:
            raise ValueError("Divider name is too long")
        return (self.start + " " + name + " ").ljust(self.length, self.fill)

    def match(self, search_str):
        """Split an input string on headers, while keeping the header name."""
        pat = "^(?=.{{{:d}}}$){} (.*) {}+$".format(self.length, self.start, self.fill)
        return re.split(pat,search_str,flags=re.M)[1:]


# TODO: perhaps subclass collections.abc.MutableMapping - however, this may not preserve order
# TODO: or, keep dict structure but remove the _order - dict has guaranteed order as of python 3.7
class Document:
    """A custom method that emulates the python 'dict', but with different construction methods and an inherent order in the keys.
    Also has string methods to appear as a proper block-based document

    sub_list: a dict of possible substitutions to make
    defaults: a dict of fallback values when passed a 'Falsey' value for a block
    div_func: the divider used when printing the Document
    buf: number of newlines to place at end of printing block
    """
    def __init__(self, contents, sub_list={}, defaults={}, div_func=None, buf=0):
        self.div = div_func
        self.subs = sub_list
        self._blocks = contents
        self._order = [] # order matters here!
        self.buf=buf
        self.defaults = defaults

    def __repr__(self):
        return "Blocks:\n"+repr(self._blocks) + "\nOrder:\n" + repr(self._order)

    def __str__(self):
        output = ""
        for block in self._order:
            if self.div:
                output += self.div.gen(block) + "\n"
            output += self._blocks[block] + "\n"*(self.buf+1)
        return output 

    def write(self,path):
        """Write to a document"""
        path.write_text(str(self))

    def __getitem__(self,bname):
        return self._blocks[bname]

    def get(self, bname, rep=""):
        if bname in self._order:
            return self._blocks[bname]
        else:
            return rep

    def __contains__(self,bname):
        return bname in self._blocks

    def __setitem__(self, bname, cstr):
        # can input blank cstr in any 'False' format
        if not cstr:
            cstr = self.defaults.get(bname,"")

        # remove trailing whitespace, starting and ending blank lines
        cstr = cstr.strip()
        
        # substitute matches in cstr with sub_list
        repl_match = lambda x: r"<\+" + str(x) + r"\+>"
        for k in self.subs.keys():
            cstr = re.sub(repl_match(k), str(self.subs[k]), cstr)

        # add _blocks, blocks; overwrites
        self._blocks[bname] = cstr
        if bname not in self._order:
            self._order.append(bname)
    
    def __missing__(self,bname):
        return ""

    def __delitem__(self, bname):
        del self._blocks[bname]
        self._order.remove(bname)


# parse the file for errors
# TODO: this is garbage, fix it (that's probably a lot of work unfortunately)
# TODO: does not catch warnings
def parse_errors(path):
    """Parse the file path for errors."""
    dct = {'errors':[],'warnings':[],'fatal':[]}
    fl = path.read_text().split("\n")

    append = False
    temp = ""
    for l in fl:
        if l.startswith("! "):
            dct['fatal'] += [l]
        if l.startswith("./.workspace/test.tex:"):
            temp = l
            append = True
        if append:
            temp += l
        if l.startswith("l."):
            temp += l
            append = False
            dct['errors'] += [temp]
    for key in ['errors','warnings','fatal']:
        if dct[key]:
            return dct
    return {}


class TexnewDocument(Document):
    """A special class of type Document with custom loading and checking types, along with a LaTeX style block delimiter"""
    def __init__(self, contents, sub_list={},defaults={}):
        # create default settings when inputting block (if block is none)
        new_defs = {
            'header':("% Template created by texnew (author: Alex Rutar); info can be found at 'https://github.com/alexrutar/texnew'.\n"
                      "% version ({})".format(__version__)),
            'file-specific preamble': "% REPLACE",
            'document start': "REPLACE\n\\end{document}"
        }
        super().__init__(contents, sub_list, div_func=Divider("%","-"), defaults={**new_defs,**defaults}, buf=2)

    # loads a file and appends blocks to current block list
    # TODO: Path, keep this as list, implement read_file with string from path object;
    # just wrap the string object with splitting to get a list, and a yaml read to get a yaml
    @classmethod
    def load(cls,fpath):
        """Constructor for TexnewDocument from filepath"""
        fl = fpath.read_text()

        # check version
        res = re.compile(r"% version \((.*)\)").search(fl)
        if not res or res.group(1).startswith("0"):
            raise FutureWarning("File version is too old!")

        # set blocks
        blocks = Divider("%","-").match(fl)
        dct = {blocks[i]:blocks[i+1] for i in range(0,len(blocks),2)}
        return cls(dct)

    def verify(self):
        """Compile and parse log file for errors."""
        self.write(RPath.workspace() / 'test.tex')

        # compile the template
        lmk_args = [
                'latexmk',
                '-pdf',
                '-interaction=nonstopmode',
                '-outdir={}'.format(RPath.workspace()),
                RPath.workspace() / 'test.tex']
        try:
            subprocess.check_output(lmk_args, stderr=subprocess.STDOUT)

            self.logfile = RPath.workspace() / 'test.log'
            self.errors = parse_errors(RPath.workspace() / 'test.log')
        except subprocess.CalledProcessError as e:
            self.errors = {'latexmk': e.output.decode()}
        clean_workspace()
        return self.errors
